"""
Methods for gathering data.

Gathers information about all puppet modules, including which are
present in our environment, their metadata, and their output of
``puppet strings``.
"""

from dataclasses import dataclass
from typing import (
    Any,
    Optional,
)
import json
import os.path
import hashlib
from glob import glob
from .puppet.strings import puppet_strings, PuppetStrings
from .cache import Cache
import logging


logger = logging.getLogger(__name__)


@dataclass
class ModuleEntry:
    """
    One entry in a module.

    :param name:
        Local name of the module, should always be the basename of path
    :param path:
        Absolute path in the filesystem where the module can be found.
    :param strings_output:
        Output of ``puppet strings``.
    :param metadata:
        Parsed contents of the modules ``metadata.json``.
    :param doc_files:
        List of absolute paths to documentation files provided by the
        upstream module.
    """

    name: str
    path: str
    strings_output: Optional[PuppetStrings]
    metadata: dict[str, Any]
    doc_files: list[str]

    def file(self, path: str) -> str:
        """Return the absolute path of a path inside the module."""
        return os.path.join(self.path, path)


def get_puppet_strings(cache: Cache, path: str) -> Optional[PuppetStrings]:
    """
    Run puppet string, but check cache first.

    The cache uses the contents of metadata.json as its key,
    so any updates without an updated metadata.json wont't be
    detected.

    Hashing the entire contents of the module was tested, but was to
    slow.
    """
    try:
        with open(os.path.join(path, 'metadata.json'), 'rb') as f:
            data = f.read()
            key = 'puppet-strings' + hashlib.sha1(data).hexdigest()
            if parsed := cache.get(key):
                result = parsed
            else:
                result = puppet_strings(path)
                cache.put(key, result)
            logger.info('Deserializing %s', path)
            return PuppetStrings.from_json(json.loads(result))
    except FileNotFoundError:
        # TODO actually run puppet strings again.
        # This is just since without a metadata.json we always get a
        # cache miss, which is slow.
        logger.info("Running 'puppet-strings %s'",
                    os.path.basename(path))
        result = puppet_strings(path)
        return PuppetStrings.from_json(json.loads(result))
        return None

        # try:
        #     with open(module.file('.git/FETCH_HEAD')) as f:
        #         st = os.stat(f.fileno())
        #         st.st_mtime
        # except FileNotFoundError:
        #     pass


def get_module(cache: Cache,
               path: str) -> ModuleEntry:
    """
    Return the metadata of a given module.

    :param cache:
        Cache objcet for modules, see python module configuration.

    :param path:
        Path of the given module.
    """
    name = os.path.basename(path)
    strings_data = get_puppet_strings(cache, path)

    try:
        with open(os.path.join(path, 'metadata.json')) as f:
            metadata = json.load(f)
    except FileNotFoundError:
        metadata = {}

    doc_files = glob(os.path.join(os.path.abspath(path), '*.md')) \
        + glob(os.path.join(os.path.abspath(path), 'LICENSE'))

    return ModuleEntry(name=name,
                       path=path,
                       strings_output=strings_data,
                       metadata=metadata,
                       doc_files=doc_files)


def get_modules(cache: Cache, dir: str) -> list[ModuleEntry]:
    """
    Return all modules present in a given directory.

    The directory should be the modules subdirectory of an environment,
    e.g. /etc/puppetlabs/code/environments/production/modules.
    """
    modules: list[ModuleEntry] = []

    for entry in sorted(list(os.scandir(dir)), key=lambda d: d.name):
        # TODO Logging
        # print('- entry', entry, file=sys.stderr)
        path = os.path.join(dir, entry)

        modules.append(get_module(cache, path))

    return modules
