"""
Functions for actually generating output.

Both generates output strings, and writes them to disk.
"""

import os
import os.path
import pathlib
# import json
import html
from .gather import ModuleEntry
from jinja2 import (
    Environment,
    FileSystemLoader,
)
# from .lookup import lookup, Ref
from .markdown import markdown
from .format import (
    format_class,
    format_type_alias,
)
from typing import (
    Optional,
    Protocol,
)
from .util import group_by
from .puppet.strings import (
    isprivate,
    PuppetStrings,
    ResourceType,
    DefinedType,
    DataTypeAlias,
    PuppetClass,
)
import logging

from .breadcrumbs import breadcrumbs
from .syntax_highlight import highlight
from dataclasses import dataclass, field
from . import templates


logger = logging.getLogger(__name__)


# TODO replace 'output' with base, or put this somewhere else
pathlib.Path('output').mkdir(exist_ok=True)
jinja = Environment(
    loader=FileSystemLoader('templates'),
    autoescape=False,
)


class HtmlSerializable(Protocol):
    """Classes which can be serialized as HTML."""

    def to_html(self) -> str:
        """Return HTML string."""
        ...

    def to_html_list(self) -> str:
        """Return HTML suitable for a list."""
        ...


@dataclass
class ResourceTypeOutput:
    """Basic HTML implementation."""

    title: str
    children: list['HtmlSerializable'] = field(default_factory=list)
    link: Optional[str] = None
    summary: Optional[str] = None

    def base(self) -> str:
        """
        Return base text of the node.

        If the node has a link, create a hyperlink, otherwise return
        it's title directly.
        """
        if self.link:
            return f'<a href="{ self.link }">{ self.title }</a>'
        else:
            return self.title

    def to_html(self) -> str:
        """Return HTML string."""
        # self.__class__.__name__
        return jinja \
            .get_template('snippets/ResourceType-index-entry.html') \
            .render(item=self)

    def to_html_list(self) -> str:
        """Return HTML suitable for a list."""
        return jinja \
            .get_template('snippets/ResourceType-list-entry.html') \
            .render(item=self)


def setup_index(base: str, modules: list[ModuleEntry], *, path_base: str) -> None:
    """
    Create the main index.html file.

    :param base:
        Path to output directory.
    :param modules:
        Modules to include in the index.
    :param path_base:
        Web path where this module will be deployed
    """
    with open(os.path.join(base, 'index.html'), 'w') as f:
        f.write(templates.index(modules=modules,
                                path_base=path_base))


@dataclass
class IndexItem:
    """
    A concrete type, on an index page.

    This will be something like a class or resource type.

    :param name:
        Name of the resource or similar
    :param file:
        Relative path to the resource.
    :param summary:
        One line summary of the resource, will be displayed in the UI.
    """

    name: str
    file: str
    summary: Optional[str] = None

    def base(self) -> str:
        """Return link to self."""
        return f'<a href="{self.file}">{ self.name }</a>'

    def to_html(self) -> str:
        """Convert item to an HTML string."""
        out: str = ''
        out += f'<dt>{self.base()}</dt>'

        if self.summary:
            out += f"<dd>{self.summary}</dd>"

        return out

    def to_html_list(self) -> str:
        """Convert itom to an HTML string sutibale for a list."""
        out: str = ''
        out += f'<li>{self.base()}</li>'
        return out


@dataclass
class IndexSubcategory:
    """
    Subheading on index page.

    Will most likely be 'Public' or 'Private' objects for the given
    top heading.
    """

    title: str
    list: list[IndexItem]

    def to_html(self) -> str:
        """Convert subcategory to an HTML string."""
        out: str = ''
        out += f'<h3>{html.escape(self.title)}</h3><dl class="overview-list">'

        for item in self.list:
            out += item.to_html()

        out += '</dl>'

        return out

    def to_html_list(self) -> str:
        """Convert itom to an html string suitable for a list."""
        pass
        out: str = ''
        out += f'<li>{html.escape(self.title)}<ul>'
        for item in self.list:
            out += item.to_html_list()
        out += '</ul></li>'

        return out


@dataclass
class IndexCategory:
    """
    Top level heading in index.

    This should be something like 'Classes', 'Types', ...
    Each entry contains a set of subentries, which can either be
    distinct entries, or sometimes subcategories (such as 'Public' and
    'Private').
    """

    title: str
    list: list[IndexSubcategory]

    def base(self) -> str:
        """Return HTML escaped title."""
        return html.escape(self.title)

    def to_html(self) -> str:
        """Return class as an HTML string."""
        out: str = ''
        out += f'<h2>{self.base()}</h2>'

        for item in self.list:
            out += item.to_html()

        return out

    def to_html_list(self) -> str:
        """Return class as an HTML string suitable for a list."""
        out: str = ''
        out += f'<li>{self.base()}<ul>'
        for item in self.list:
            out += item.to_html_list()
        out += '</ul></li>'

        return out


def index_item(obj: PuppetClass | DefinedType) -> IndexItem:
    """
    Format a puppet type declaration into an index entry.

    :param obj:
        A dictionary at least containing the keys 'name' and 'file',
        and optionally containing 'docstring'. If docstring is present
        then a summary tag is searched for, and added to the resulting
        object.
    """
    name = obj.name

    out: IndexItem = IndexItem(
        file=os.path.splitext(obj.file)[0],
        name=name,
    )

    for tag in obj.docstring.tags:
        if tag.tag_name == 'summary':
            out.summary = markdown(tag.text)
            break

    return out


def class_index(class_list: list[PuppetClass]) -> IndexCategory:
    """Prepage class index list."""
    groups = group_by(isprivate, class_list)

    lst: list[IndexSubcategory] = []

    if publics := groups.get(False):
        lst.append(IndexSubcategory(
            title='Public Classes',
            list=[index_item(i) for i in publics],
        ))

    if privates := groups.get(True):
        lst.append(IndexSubcategory(
            title='Private Classes',
            list=[index_item(i) for i in privates],
        ))

    return IndexCategory(
        title='Classes',
        list=lst
    )


def defined_types_index(defined_list: list[DefinedType]) -> IndexCategory:
    """
    Prepare defined types index list.

    These are puppet types introduces by puppet code.
    Each only has one implemenattion.
    """
    groups = group_by(isprivate, defined_list)

    lst: list[IndexSubcategory] = []

    if publics := groups.get(False):
        lst.append(IndexSubcategory(
            title='Public Defined Types',
            list=[index_item(i) for i in publics],
        ))

    if privates := groups.get(True):
        lst.append(IndexSubcategory(
            title='Private Defined Types',
            list=[index_item(i) for i in privates],
        ))

    return IndexCategory(
        title='Defined Types',
        list=lst
    )


def type_aliases_index(alias_list: list[DataTypeAlias]) -> IndexCategory:
    """Prepare type alias index list."""
    groups = group_by(isprivate, alias_list)
    lst: list[IndexSubcategory] = []
    if publics := groups.get(False):
        lst.append(IndexSubcategory(
            title='Public Type Aliases',
            list=[IndexItem(name=i.name,
                            file=os.path.splitext(i.file)[0])
                  for i in publics],
        ))

    if privates := groups.get(True):
        lst.append(IndexSubcategory(
            title='Private Type Aliases',
            list=[IndexItem(name=i.name,
                            file=os.path.splitext(i.file)[0])
                  for i in privates],
        ))

    return IndexCategory(
        title='Type Aliases',
        list=lst,
    )


@dataclass
class ResourceIndex:
    """Placeholder."""

    title: str
    children: list[HtmlSerializable]

    def to_html(self) -> str:
        """Return something."""
        out: str = ''
        out += f'<h2>{self.title}</h2>'
        out += '<dl>'
        for child in self.children:
            out += child.to_html()
        out += '</dl>'
        return out

    def to_html_list(self) -> str:
        """Return something."""
        out: str = ''
        out += f'<li>{self.title}<ul>'
        for child in self.children:
            out += child.to_html_list()
        out += '</ul></li>'
        return out


def resource_type_index(resource_types: list[ResourceType]) -> list[HtmlSerializable]:
    """Generate index for all known resource types."""
    lst: list[HtmlSerializable] = []

    for resource_type in resource_types:
        # resource_type['file']
        # resource_type['docstring']

        items: list[HtmlSerializable] = []
        if providers := resource_type.providers:
            for provider in providers:
                # TODO summary tag?
                # TODO, instead, render markdown and take first paragraph
                documentation = provider.docstring.text.split('\n')[0]
                items.append(ResourceTypeOutput(
                    title=provider.name,
                    link=provider.file,
                    summary=documentation))

        lst.append(ResourceTypeOutput(title=resource_type.name,
                                      children=items))

    return lst


# def resource_types_index(resource_list: list) -> IndexCategory:
#     """
#     Prepare resource type index list.
#
#     These are the resource types introduced through ruby. Each can
#     have multiple implementations.
#     """
#     return {}


def setup_module_index(*,
                       base: str,
                       module: ModuleEntry,
                       data: PuppetStrings,
                       path_base: str,
                       doc_files: dict[str, str],
                       ) -> None:
    """Create the index file for a specific module."""
    content: list[ResourceIndex | IndexCategory] = []

    content.append(class_index(data.puppet_classes))

    # data['data_types']
    content.append(IndexCategory(
        title='Data types not yet implmented',
        list=[],
    ))

    content.append(type_aliases_index(data.data_type_aliases))

    content.append(defined_types_index(data.defined_types))

    content.append(ResourceIndex(
        title='Resource Types',
        children=resource_type_index(data.resource_types)))

    # data['providers']
    content.append(IndexCategory(
        title='Providers not yet implmented',
        list=[],
    ))

    # data['puppet_functions']
    content.append(IndexCategory(
        title='Puppet Functions not yet implmented',
        list=[],
    ))

    # templates/
    # files/
    # examples or tests/
    # (spec)/
    # lib/puppet_x/
    # lib/facter/
    # facts.d/
    # data/
    # hiera.yaml

    # data['puppet_tasks']
    content.append(IndexCategory(
        title='Puppet Tasks not yet implmented',
        list=[],
    ))

    # data['puppet_plans']
    content.append(IndexCategory(
        title='Puppet Plans not yet implmented',
        list=[],
    ))

    crumbs = breadcrumbs(
            ('Environment', ''),
            module.name,
            )

    with open(os.path.join(base, 'index.html'), 'w') as f:
        f.write(templates.module_index(
            module_name=module.name,
            module_author='TODO',  # module.metadata['author'],
            breadcrumbs=crumbs,
            content=content,
            path_base=path_base,
            doc_files=list(doc_files.items())))


GENERATED_MESSAGE = '<!-- DO NOT EDIT: This document was generated by Puppet Strings -->\n'
"""
REFERENCE.md files generated by Puppet Strings include this string on
their third line. We use this to ignore auto-generated files, since we
replace that output.
"""


def setup_module(base: str, module: ModuleEntry, *, path_base: str) -> None:
    """
    Create all output files for a puppet module.

    Will generate a directory under base for the module.

    :param base:
        Path to output directory.
    :param modules:
        Modules to document.
    :param path_base:
        Web path where this module will be deployed
    """
    path = os.path.join(base, module.name)
    pathlib.Path(path).mkdir(exist_ok=True)
    if not module.strings_output:
        logger.warning("No strings output for %s", module.name)
        return

    data = module.strings_output

    for puppet_class in data.puppet_classes + data.defined_types:
        logger.info('Formamting %s', puppet_class.name)
        # localpath = puppet_class['name'].split('::')
        localpath, _ = os.path.splitext(puppet_class.file)
        dir = os.path.join(path, localpath)
        pathlib.Path(dir).mkdir(parents=True, exist_ok=True)
        # puppet_class['docstring']
        # puppet_class['defaults']

        # TODO option to add .txt extension (for web serverse which
        # treat .pp as application/binary)
        with open(os.path.join(dir, 'source.pp.txt'), 'wb') as f:
            with open(module.file(puppet_class.file), 'rb') as g:
                f.write(g.read())

        crumbs = breadcrumbs(
                ('Environment', ''),
                module.name,
                (puppet_class.name,
                 'manifests/' + '/'.join(puppet_class.name.split('::')[1:])),
                'This',
                )

        with open(os.path.join(dir, 'source.pp.html'), 'w') as f:

            with open(module.file(puppet_class.file), 'r') as g:
                f.write(templates.code_page(
                    title='',
                    content=highlight(g.read(), 'puppet'),
                    path_base=path_base,
                    breadcrumbs=crumbs))

        # TODO reimplement this?
        # with open(os.path.join(dir, 'source.json'), 'w') as f:
        #     json.dump(puppet_class, f, indent=2)

        # with open(os.path.join(dir, 'source.pp.html'), 'w') as f:
        #     f.write(format_class(puppet_class))

        crumbs = breadcrumbs(
                ('Environment', ''),
                module.name,
                (puppet_class.name,
                 'manifests/' + '/'.join(puppet_class.name.split('::')[1:])),
                )

        title, body = format_class(puppet_class)
        with open(os.path.join(dir, 'index.html'), 'w') as f:
            f.write(templates.code_page(
                title=title,
                content=body,
                path_base=path_base,
                breadcrumbs=crumbs))

        # puppet_class['file']
        # puppet_class['line']

    for type_alias in data.data_type_aliases:
        logger.debug('Formamting %s', type_alias.name)

        localpath, _ = os.path.splitext(type_alias.file)
        dir = os.path.join(path, localpath)
        pathlib.Path(dir).mkdir(parents=True, exist_ok=True)

        with open(os.path.join(dir, 'source.pp.txt'), 'w') as f:
            f.write(type_alias.alias_of)

        # TODO reimplement this?
        # with open(os.path.join(dir, 'source.json'), 'w') as f:
        #     json.dump(type_alias, f, indent=2)

        title, body = format_type_alias(type_alias, type_alias.name)
        with open(os.path.join(dir, 'index.html'), 'w') as f:
            f.write(templates.code_page(
                title=title,
                content=body,
                path_base=path_base))

    # data['data_type_aliases']
    # data['defined_types']
    # data['resource_types']

    files: dict[str, str] = {}
    for file in module.doc_files:
        logger.debug('Formamting %s', file)

        basename = os.path.basename(file)
        if basename == 'REFERENCE.md':
            with open(file) as f:
                f.readline()
                f.readline()
                line3 = f.readline()
                if line3 == GENERATED_MESSAGE:
                    continue
        files[basename] = file

    doc_files: dict[str, str] = {}
    for filename, filepath in files.items():
        logger.debug('Formamting %s', filename)

        name, _ = os.path.splitext(filename)
        with open(filepath) as f:
            raw_content = f.read()

        pathlib.Path(os.path.join(path, name)).mkdir(exist_ok=True)
        out_path = os.path.join(path, name, 'index.html')

        if filename.endswith('.md'):
            content = markdown(raw_content)
        else:
            content = '<pre>' + html.escape(raw_content) + '</pre>'

        crumbs = breadcrumbs(('Environment', ''),
                             module.name,
                             name)

        with open(out_path, 'w') as f:
            f.write(templates.content(
                content=content,
                path_base=path_base,
                breadcrumbs=crumbs))

        doc_files[name] = os.path.join(module.name, name, 'index.html')

    setup_module_index(base=path,
                       module=module,
                       data=data,
                       path_base=path_base,
                       doc_files=doc_files,
                       )
