#!/usr/bin/env python3

"""
(S)CSS generators for help with highlighting.

:ref:`highlight-readme`.
"""

import yaml
from datetime import datetime
import argparse

CSS_RULES = '''
    color:            var(--hl-{name}-color);
    background-color: var(--hl-{name}-background);
    font-style:       var(--hl-{name}-font-style);
    font-weight:      var(--hl-{name}-font-weight);
'''

HEADER = '''/*
* File autogenerated from {source}.
* Generated at {time}
*/
'''

HIGHLIGHT_GROUPS = [
    "comment",
    "error",
    "escape",
    "interpolate",
    "keyword",
    "line",
    "number",
    "operator",
    "regex",
    "special",
    "string",
    "variable",
]

HIGHLIGHT_VARIABLES = [
    'color',
    'background',
    'font-style',
    'font-weight',
]


def build_highlight_map(source_file: str) -> str:
    """
    Generate CSS rules from source file.

    :param source_file:
        A yaml file containing mappings from our abstract highlight
        groups, to concrete CSS classes.
    :returns:
        Generated CSS code, applicable to be written directly to a
        file.
    """
    with open(source_file) as f:
        data = yaml.full_load(f)
    result = ''
    result += HEADER.format(source=source_file,
                            time=datetime.now())
    for name, classes in data.items():
        output: str = '\n'
        output += ', '.join(f'.{c}' for c in classes)
        output += ' {'
        output += CSS_RULES.format(name=name)
        output += '}'

        result += output + '\n'

    return result


def build_colorscheme(source_file: str,
                      selector: str = ':root') -> str:
    """
    Build CSS colorscheme from source file.

    :param selector:
        CSS selector to use for this group.
    :param source_file:
        Yaml file containing mapping from abrtract highlighting groups,
        to CSS attributes.
    :returns:
        Generated CSS code, applicable to be written directly to a
        file.
    """
    with open(source_file) as f:
        data = yaml.full_load(f)

    result = ''
    result += HEADER.format(source=source_file,
                            time=datetime.now())
    result += f'{selector} {{\n'
    for group_key in HIGHLIGHT_GROUPS:
        if group := data.get(group_key):
            for value_key in HIGHLIGHT_VARIABLES:
                if value := group.get(value_key):
                    result += f'    --hl-{group_key}-{value_key}: {value};\n'
            if raw := group.get('raw-append'):
                result += raw + '\n'
    # Raw CSS rules, useful to capture something not handled by
    # the abstract groups.
    if raw := data.get('raw-append'):
        result += raw + '\n'

    result += '}\n'
    return result


def run_colorscheme(args: argparse.Namespace) -> None:
    """Entry point when using the colorscheme mode."""
    print(build_colorscheme(selector=args.selector,
                            source_file=args.source))


def run_highlight(args: argparse.Namespace) -> None:
    """Entry point when using the highlight mode."""
    print(build_highlight_map(source_file=args.source))


def main() -> None:
    """Primary entry point of program."""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    parser_hl = subparsers.add_parser('highlight')
    parser_hl.set_defaults(func=run_highlight)
    parser_hl.add_argument('source')

    parser_cs = subparsers.add_parser('colorscheme')
    parser_cs.set_defaults(func=run_colorscheme)
    parser_cs.add_argument('--selector',
                           default=':root')
    parser_cs.add_argument('source')

    args = parser.parse_args()

    args.func(args)


if __name__ == '__main__':
    main()
